//
//  main.m
//  EndpointSecurity_TCC
//
//  Created by Patrick Wardle on 3/20/25.
//

#import <bsm/libbsm.h>
#import <Cocoa/Cocoa.h>
#import <EndpointSecurity/EndpointSecurity.h>

es_client_t* client = NULL;

BOOL monitorTCC(void);
void printESClientError(es_new_client_result_t result);

@interface NSString (ESToken)
+ (instancetype)stringWithESToken:(es_string_token_t)token;
@end

@implementation NSString (ESToken)

+ (instancetype)stringWithESToken:(es_string_token_t)token {
    if (token.length == 0 || token.data == NULL) {
        return @"";
    }
    return [[NSString alloc] initWithBytes:token.data
                                    length:token.length
                                  encoding:NSUTF8StringEncoding];
}
@end

static NSString *NSStringFromTCCIdentityType(es_tcc_identity_type_t type) {
    switch (type) {
        case ES_TCC_IDENTITY_TYPE_BUNDLE_ID:
            return @"Bundle ID";
        case ES_TCC_IDENTITY_TYPE_EXECUTABLE_PATH:
            return @"Executable Path";
        case ES_TCC_IDENTITY_TYPE_POLICY_ID:
            return @"Policy ID";
        case ES_TCC_IDENTITY_TYPE_FILE_PROVIDER_DOMAIN_ID:
            return @"File Provider Domain ID";
        default:
            return [NSString stringWithFormat:@"Unknown (%d)", type];
    }
}

static NSString *NSStringFromTCCEventType(es_tcc_event_type_t type) {
    switch (type) {
        case ES_TCC_EVENT_TYPE_UNKNOWN:
            return @"Unknown";
        case ES_TCC_EVENT_TYPE_CREATE:
            return @"Create";
        case ES_TCC_EVENT_TYPE_MODIFY:
            return @"Modify";
        case ES_TCC_EVENT_TYPE_DELETE:
            return @"Delete";
        default:
            return [NSString stringWithFormat:@"Unknown (%d)", type];
    }
}

static NSString *NSStringFromTCCAuthorizationRight(es_tcc_authorization_right_t right) {
    switch (right) {
        case ES_TCC_AUTHORIZATION_RIGHT_DENIED:
            return @"Denied";
        case ES_TCC_AUTHORIZATION_RIGHT_UNKNOWN:
            return @"Unknown";
        case ES_TCC_AUTHORIZATION_RIGHT_ALLOWED:
            return @"Allowed";
        case ES_TCC_AUTHORIZATION_RIGHT_LIMITED:
            return @"Limited";
        case ES_TCC_AUTHORIZATION_RIGHT_ADD_MODIFY_ADDED:
            return @"Add/Modify (Added)";
        case ES_TCC_AUTHORIZATION_RIGHT_SESSION_PID:
            return @"Session PID";
        case ES_TCC_AUTHORIZATION_RIGHT_LEARN_MORE:
            return @"Learn More";
        default:
            return [NSString stringWithFormat:@"Unknown (%d)", right];
    }
}

static NSString *NSStringFromTCCAuthorizationReason(es_tcc_authorization_reason_t reason) {
    switch (reason) {
        case ES_TCC_AUTHORIZATION_REASON_NONE:
            return @"None";
        case ES_TCC_AUTHORIZATION_REASON_ERROR:
            return @"Error";
        case ES_TCC_AUTHORIZATION_REASON_USER_CONSENT:
            return @"User Consent";
        case ES_TCC_AUTHORIZATION_REASON_USER_SET:
            return @"User Set (via Preferences)";
        case ES_TCC_AUTHORIZATION_REASON_SYSTEM_SET:
            return @"System Set";
        case ES_TCC_AUTHORIZATION_REASON_SERVICE_POLICY:
            return @"Service Policy";
        case ES_TCC_AUTHORIZATION_REASON_MDM_POLICY:
            return @"MDM Policy";
        case ES_TCC_AUTHORIZATION_REASON_SERVICE_OVERRIDE_POLICY:
            return @"Service Override Policy";
        case ES_TCC_AUTHORIZATION_REASON_MISSING_USAGE_STRING:
            return @"Missing Usage String";
        case ES_TCC_AUTHORIZATION_REASON_PROMPT_TIMEOUT:
            return @"Prompt Timeout";
        case ES_TCC_AUTHORIZATION_REASON_PREFLIGHT_UNKNOWN:
            return @"Preflight Unknown";
        case ES_TCC_AUTHORIZATION_REASON_ENTITLED:
            return @"Entitled";
        case ES_TCC_AUTHORIZATION_REASON_APP_TYPE_POLICY:
            return @"App Type Policy";
        case ES_TCC_AUTHORIZATION_REASON_PROMPT_CANCEL:
            return @"Prompt Cancel";
        default:
            return [NSString stringWithFormat:@"Unknown (%d)", reason];
    }
}

BOOL monitorTCC(void)
{
    BOOL started = NO;
    es_new_client_result_t result = 0;
    es_event_type_t events[] = {ES_EVENT_TYPE_NOTIFY_TCC_MODIFY};
   
    result = es_new_client(&client, ^(es_client_t *client, const es_message_t *message) {
        
        if(message->event_type != ES_EVENT_TYPE_NOTIFY_TCC_MODIFY) return;
        
        es_process_t* process = message->process;
        es_event_tcc_modify_t* event = message->event.tcc_modify;
        
        NSLog(@"received 'ES_EVENT_TYPE_NOTIFY_TCC_MODIFY' message");
        NSLog(@"process: %d:%@", audit_token_to_pid(process->audit_token), [NSString stringWithESToken:process->executable->path]);
        
        NSString* service = [NSString stringWithESToken:event->service];
        NSString* identity = [NSString stringWithESToken:event->identity];
        
        NSLog(@"service: %@", service);
        NSLog(@"identity: %@", identity);
        
        NSLog(@"identity type: %@", NSStringFromTCCIdentityType(event->identity_type));
        NSLog(@"update type: %@", NSStringFromTCCEventType(event->update_type));
        
        if(NULL != event->instigator) {
            NSLog(@"instigator process: %@", [NSString stringWithESToken:event->instigator->executable->path]);
        }
        else {
            NSLog(@"instigator process is NULL");
        }
        
        if(NULL != event->responsible_token) {
            NSLog(@"responsible process pid: %d", audit_token_to_pid(*event->responsible_token));
        }
        else {
            NSLog(@"responsible token is NULL");
        }
        
        if(NULL != event->responsible) {
            NSLog(@"responsible process: %@", [NSString stringWithESToken:event->responsible->executable->path]);
        }
        else {
            NSLog(@"responsible process is NULL");
        }
                       
        NSLog(@"right: %@", NSStringFromTCCAuthorizationRight(message->event.tcc_modify->right));
        NSLog(@"reason: %@", NSStringFromTCCAuthorizationReason(message->event.tcc_modify->reason));
        
    });
    if(ES_NEW_CLIENT_RESULT_SUCCESS != result)
    {
        printESClientError(result);
        goto bail;
    }
    
    if(ES_RETURN_SUCCESS != es_subscribe(client, events, sizeof(events)/sizeof(events[0])))
    {
        printf("\n\nERROR: 'es_subscribe' failed\n");
        goto bail;
    }

    started = YES;
    
bail:
    
    return started;
}

void printESClientError(es_new_client_result_t result)
{
    //err msg
    printf("ERROR: 'es_new_client()' failed with %#x\n", result);
    switch(result){
            
        //not entitled
        case ES_NEW_CLIENT_RESULT_ERR_NOT_ENTITLED:
            printf("ES_NEW_CLIENT_RESULT_ERR_NOT_ENTITLED: \"The caller is not properly entitled to connect\"\n\n");
            break;
                  
        //not permitted
        case ES_NEW_CLIENT_RESULT_ERR_NOT_PERMITTED:
            printf("ES_NEW_CLIENT_RESULT_ERR_NOT_PERMITTED: \"The caller is not permitted to connect. They lack Transparency, Consent, and Control (TCC) approval from the user.\"\n\n");
            break;
                  
        //not privileged
        case ES_NEW_CLIENT_RESULT_ERR_NOT_PRIVILEGED:
            printf("ES_NEW_CLIENT_RESULT_ERR_NOT_PRIVILEGED: \"The caller is not running as root\"\n\n");
            break;
            
        default:
            break;
    }
    
    return;
}

int main(int argc, const char * argv[]) {
    
    monitorTCC();
    [[NSRunLoop mainRunLoop] run];

}
